import os
from abc import ABC, abstractmethod
from typing import Dict

from twang.track import BaseTrack, LibrosaTrack
from twang.types import AudioFormat
from twang.util import file_util


class SourceSeparationDict(Dict[str, LibrosaTrack]):
    def save(self, save_dir: str, audio_format: AudioFormat = AudioFormat.WAV):
        file_util.mkdir(save_dir)
        for source_name, source_track in self.items():
            source_track.save(os.path.join(save_dir, f"{source_name}.{audio_format.value}"), audio_format=audio_format)


class SourceSeparation(ABC):
    """A generic base class defining the API for all source-separation implementations."""

    @abstractmethod
    def run(self, track: BaseTrack) -> SourceSeparationDict:
        """
        Runs a source separation algorithm, separating the `track` audio into different components. Note that the API
        supports basing any BaseTrack subclass (e.g. PyDubTrack, LibrosaTrack). It's up to the SourceSeparation
        subclass implementation to ensure that this is possible.
        """
        raise NotImplementedError



import nussl

from twang.source_separation.base import SourceSeparation, SourceSeparationDict
from twang.track import BaseTrack, LibrosaTrack


def _track_to_nussl_audio_signal(track: BaseTrack) -> nussl.AudioSignal:
    librosa_track = track.to_librosa_track()
    return nussl.AudioSignal(audio_data_array=librosa_track.y, sample_rate=librosa_track.sr)


def _audio_signal_to_librosa_track(audio_signal: nussl.AudioSignal) -> LibrosaTrack:
    return LibrosaTrack(audio_signal.audio_data, sr=audio_signal.sample_rate)


class Repet(SourceSeparation):
    """https://nussl.github.io/docs/separation.html#foreground-background-via-repet"""

    repet_cls = nussl.separation.primitive.Repet

    def run(self, track: BaseTrack) -> SourceSeparationDict:
        audio_signal = _track_to_nussl_audio_signal(track)
        repet = self.repet_cls(audio_signal)
        estimates = repet()
        return {
            "background": _audio_signal_to_librosa_track(estimates[0]),
            "foreground": _audio_signal_to_librosa_track(estimates[1]),
        }


class RepetSim(Repet):
    """https://nussl.github.io/docs/separation.html#foreground-background-via-repet-sim"""

    repet_cls = nussl.separation.primitive.RepetSim



import enum
from typing import List, Mapping, Union


class AudioFormat(enum.Enum):
    NONE = None
    WAV = "wav"
    OGG = "ogg"
    FLAC = "flac"
    MP3 = "mp3"


JSON = Union[str, int, float, bool, None, Mapping[str, "JSON"], List["JSON"]]



from abc import ABC, abstractmethod
from typing import Any, Optional, Type, TypeVar, Union

import librosa
import numpy as np
import pydub
import soundfile as sf
from IPython import display as ipd

from twang.types import AudioFormat
from twang.util import time as time_util

_BaseTrack = TypeVar("_BaseTrack", bound="BaseTrack")


class BaseTrack(ABC):
    """Defines a standard API for working with audio tracks."""

    # The audio signal; its format depends on the `Track` implementation.
    y: Any

    # Sampling rate
    sr: int

    # If the track exists on disk, this attribute should store its file path
    file_path: Optional[str] = None

    @classmethod
    @abstractmethod
    def from_file(cls: Type[_BaseTrack], file_path: str) -> _BaseTrack:
        """Load an audio file from disk."""
        ...

    @abstractmethod
    def _repr_html_(self) -> str:
        """Display the BaseTrack in a Jupyter notebook"""
        ...

    @abstractmethod
    def __getitem__(self, ms_or_slice: Union[int, slice]):
        """A standardised API for slicing the audio into a shorter `BaseTrack`,  or accessing a specific timestamp."""
        ...

    @abstractmethod
    def save(self, save_path: str, audio_format: AudioFormat = AudioFormat.WAV):
        """Serialize the track to disk.

        Args:
            save_path: path at which to save the audio file
            audio_format: format in which to save the track
            # TODO: add more customisation potential here
        """
        ...

    @abstractmethod
    def to_librosa_track(self) -> "LibrosaTrack":
        """Convert this track to a LibrosaTrack implementation, the implementation used by almost all modules."""
        ...

    @property
    @abstractmethod
    def duration(self) -> float:
        ...

    @property
    @abstractmethod
    def bpm(self) -> float:
        ...


class PyDubTrack(BaseTrack):
    """# TODO: docstring"""

    y: pydub.AudioSegment

    def __init__(self, y: pydub.AudioSegment, file_path: Optional[str] = None):
        self.y = y
        self.sr = y.frame_rate
        self.file_path = file_path

    @classmethod
    def from_file(cls, file_path: str, audio_format: AudioFormat = AudioFormat.NONE):
        audio_segment = pydub.AudioSegment.from_file(file_path, format=audio_format.value)
        return cls(y=audio_segment, file_path=file_path)

    @classmethod
    def from_file_snippet(
        cls,
        file_path: str,
        start_time: float,
        end_time: float,
        audio_format: AudioFormat = AudioFormat.NONE,
    ):
        track = cls.from_file(file_path, audio_format=audio_format)
        return track[int(time_util.s_to_ms(start_time)) : int(time_util.s_to_ms(end_time))]

    @classmethod
    def from_librosa_track(cls, librosa_track: "LibrosaTrack"):
        y = y_new = librosa_track.y

        # interleave left & right channels into a 1D array
        if len(y.shape) == 2:
            y_new = np.zeros((y.shape[1] * 2), dtype=y.dtype)
            y_new[::2] = y[0]
            y_new[1::2] = y[1]

        return cls(
            y=pydub.AudioSegment(
                y_new.tobytes(),
                frame_rate=librosa_track.sr,
                sample_width=y_new.dtype.itemsize,
                channels=len(y.shape),
            ),
            file_path=librosa_track.file_path,
        )

    def to_librosa_track(self) -> "LibrosaTrack":
        return LibrosaTrack.from_pydub_track(self)

    def _repr_html_(self) -> str:
        return self.y._repr_html_()

    def __getitem__(self, ms_or_slice: Union[int, slice]):
        """TODO: test & document"""
        self.file_path = None  # because the object no longer corresponds to the file on disk
        self.y = self.y[ms_or_slice]
        return self

    def save(self, save_path: str, audio_format: AudioFormat = AudioFormat.WAV):
        self.y.export(save_path, format=audio_format.value)

    @property
    def duration(self) -> float:
        raise NotImplementedError

    @property
    def bpm(self) -> float:
        raise NotImplementedError


class LibrosaTrack(BaseTrack):
    """TODO: docstring"""

    y: np.ndarray

    def __init__(self, y: np.ndarray, sr: int, use_onset: bool = False, file_path: Optional[str] = None):
        self.y = y
        self.sr = sr
        self.use_onset = use_onset
        self.file_path = file_path

    @classmethod
    def from_file(cls, file_path: str, use_onset: bool = False):
        return cls(*librosa.load(file_path), use_onset=use_onset, file_path=file_path)

    @classmethod
    def from_pydub_track(cls, pydub_track: PyDubTrack, use_onset: bool = False):
        samples = pydub_track.y.get_array_of_samples()

        # convert from int16 —> float32
        samples = librosa.util.buf_to_float(samples, n_bytes=2, dtype=np.float32)

        # pydub interleaves the left & right channels, => undo that to reconstruct a 2D array
        if pydub_track.y.channels == 2:
            samples = np.array([samples[::2], samples[1::2]])

        return cls(y=samples, sr=pydub_track.sr, use_onset=use_onset, file_path=pydub_track.file_path)

    def _repr_html_(self) -> str:
        return ipd.Audio(self.y, rate=self.sr)._repr_html_()

    def __getitem__(self, ms_or_slice: Union[int, slice]):
        # TODO: implement
        raise NotImplementedError

    def save(self, save_path: str, audio_format: AudioFormat = AudioFormat.WAV):
        sf.write(
            save_path,
            data=self.y.T,
            samplerate=self.sr,
            format=audio_format.name,
            subtype="vorbis" if audio_format == AudioFormat.OGG else "PCM_24",
        )

    def to_librosa_track(self) -> "LibrosaTrack":
        return self

    @property
    def duration(self) -> float:
        # TODO: add different formatting options here
        return librosa.get_duration(y=self.y)

    @property
    def bpm(self) -> float:
        raise NotImplementedError



import enum
import os
import tempfile

import torch
from demucs.apply import apply_model
from demucs.pretrained import DEFAULT_MODEL, get_model
from demucs.repo import AnyModel
from demucs.separate import load_track

from twang.source_separation.base import SourceSeparation, SourceSeparationDict
from twang.track.base import BaseTrack, LibrosaTrack


class DemucsModelType(enum.Enum):
    DEFAULT_MODEL = DEFAULT_MODEL

    # A fine-tuned version of the default that takes 4 times longer to run but is a bit better
    HTDEMUCS_FT = "htdemucs_ft"

    # Adds "piano" and "guitar" sources to the default four ["drums", "bass", "other", "vocals"].
    HTDEMUCS_6S = "htdemucs_6s"


def _track_to_demucs_audio_file(track: BaseTrack, model: AnyModel) -> torch.Tensor:
    """Converts any kind of BaseTrack to a Demucs audio file (i.e. a torch Tensor representation)."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        track_path = track.file_path or os.path.join(tmp_dir, "track.wav")
        track.save(track_path)
        demucs_audio_file = load_track(track_path, model.audio_channels, model.samplerate)
    return demucs_audio_file


class DemucsSourceSeparation(SourceSeparation):
    """https://github.com/facebookresearch/demucs."""

    def __init__(self, model: DemucsModelType = DemucsModelType.DEFAULT_MODEL, device: str | None = None):
        self.device = device or ("cuda" if torch.cuda.is_available() else "cpu")
        # TODO: allow GPU usage (i.e. obey self.device)
        self.model: AnyModel = get_model(model.value)  # .cpu().eval()

    def run(
        self,
        track: BaseTrack,
        shifts: int = 1,
        split: bool = True,
        overlap: float = 0.25,
        jobs: int = 0,
    ) -> SourceSeparationDict:
        """"""

        # Convert our track representation to a demucs `AudioFile`
        # TODO: work out what is being "mean"ed here ...
        # TODO: we're normalizing here ?
        audio_file = _track_to_demucs_audio_file(track, self.model)
        ref = audio_file.mean(0)
        audio_file = (audio_file - ref.mean()) / ref.std()

        sources = apply_model(
            self.model,
            audio_file[None],
            device=self.device,
            shifts=shifts,
            split=split,
            overlap=overlap,
            progress=True,
            num_workers=jobs,
        )[0]
        sources = sources * ref.std() + ref.mean()

        return SourceSeparationDict(
            {
                name: LibrosaTrack(y=source.numpy(), sr=self.model.samplerate)
                for source, name in zip(sources, self.model.sources)
            }
        )



def to_s(minutes: float, seconds: float) -> float:
    return (minutes * 60) + seconds


def to_ms(minutes: float, seconds: float):
    return 1000 * to_s(minutes, seconds)


def s_to_ms(seconds: float) -> float:
    return 1000 * seconds



import librosa

from twang.source_separation.base import SourceSeparation, SourceSeparationDict
from twang.track import BaseTrack, LibrosaTrack


class HarmonicPercussiveSourceSeparation(SourceSeparation):
    """Computes the short-time Fourier transform of y and decomposes it into its harmonic & percussive components."""

    def run(self, track: BaseTrack) -> SourceSeparationDict:
        D_harmonic, D_percussive = librosa.decompose.hpss(librosa.stft(track.to_librosa_track().y))
        return SourceSeparationDict(
            {
                "harmonic": LibrosaTrack(y=librosa.istft(D_harmonic), sr=track.sr),
                "percussive": LibrosaTrack(y=librosa.istft(D_percussive), sr=track.sr),
            }
        )



import pickle
from pathlib import Path
from typing import Any

import yaml
from cloudpathlib import CloudPath, GSPath

from twang.types import JSON


def _is_gcs_path(file_path: str) -> bool:
    return file_path.startswith(GSPath.cloud_prefix)


def _is_cloud_path(file_path: str) -> bool:
    return _is_gcs_path(file_path)


def _is_local_path(file_path: str) -> bool:
    return not _is_cloud_path(file_path)


def file_exists(file_path: str) -> bool:
    """Checks whether a local or cloud `file_path` 'exists' (is a valid file or directory)."""
    if _is_cloud_path(file_path):
        return CloudPath(file_path).exists()
    else:
        return Path(file_path).exists()


def mkdir(dir_path: str):
    path = CloudPath(dir_path) if _is_cloud_path(dir_path) else Path(dir_path)
    path.mkdir(parents=True, exist_ok=True)


def load_pickle(file_path: str, **kwargs) -> Any:
    """Load a pickled object from a given file_path, either local or cloud."""
    path = CloudPath(file_path) if _is_cloud_path(file_path) else Path(file_path)
    with path.open("rb") as f:
        return pickle.load(f, **kwargs)


def save_pickle(obj_to_pickle: Any, file_path: str, **kwargs) -> Any:
    """Saves an object as a pickle file to a given file_path, either local or cloud."""
    path = CloudPath(file_path) if _is_cloud_path(file_path) else Path(file_path)
    with path.open("wb") as f:
        pickle.dump(obj_to_pickle, f, **kwargs)


def load_yaml(file_path: str) -> JSON:
    """Loads a YAML file to JSON"""
    path = CloudPath(file_path) if _is_cloud_path(file_path) else Path(file_path)
    with path.open("r") as f:
        return yaml.load(stream=f, Loader=yaml.Loader)


def save_yaml(yaml_obj: JSON, file_path: str) -> None:
    """Saves a YAML object to file"""
    path = CloudPath(file_path) if _is_cloud_path(file_path) else Path(file_path)
    with path.open("w") as f:
        yaml.dump(data=yaml_obj, stream=f, Dumper=yaml.Dumper, line_break="\n")
